# Generated by Django 4.2.13 on 2024-06-25 16:40

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="NamedSerialNumber",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(blank=True, max_length=50, verbose_name="name"),
                ),
                (
                    "next_counter",
                    models.IntegerField(default=1, verbose_name="next counter"),
                ),
                ("step", models.IntegerField(default=1, verbose_name="step")),
                ("prefix", models.CharField(max_length=10, verbose_name="prefix")),
                ("width", models.IntegerField(verbose_name="number width")),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="namedserialnumber",
            constraint=models.UniqueConstraint(
                fields=("name", "content_type"), name="unique_name_for_content_type"
            ),
        ),
    ]



def sort_title(title: str) -> str:
    """Sort a title by the first word, ignoring articles."""
    articles = {"a", "an", "the"}

    title = title.lower()

    first, _, rest = title.partition(" ")
    return f"{rest}, {first}" if first in articles else title



from django.conf import settings
from django.db import models

AUTH_USER_MODEL = getattr(settings, "AUTH_USER_MODEL", "auth.User")


class AutoCreatedByField(models.ForeignKey):
    """A ForeignKey to the user model that is not editable and has a related name
    of '+'.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", False)
        kwargs.setdefault("related_name", "+")
        kwargs.setdefault("to", AUTH_USER_MODEL)
        kwargs.setdefault("on_delete", models.PROTECT)
        super().__init__(*args, **kwargs)


class AutoModifiedByField(models.ForeignKey):
    """A ForeignKey to the user model that is not editable and has a related name
    of '+'.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", False)
        kwargs.setdefault("related_name", "+")
        kwargs.setdefault("to", AUTH_USER_MODEL)
        kwargs.setdefault("on_delete", models.PROTECT)
        super().__init__(*args, **kwargs)



from model_utils.models import TimeStampedModel  # type: ignore[import-untyped]

from django_rubble.fields.auto_fields import AutoCreatedByField, AutoModifiedByField


class StampedModel(TimeStampedModel):
    """A model that has created_by and modified_by fields that are automatically
    set to the user that created or modified the instance."""

    created_by = AutoCreatedByField()
    modified_by = AutoModifiedByField()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        update_fields = kwargs.get("update_fields", None)
        if update_fields is not None:
            if "modified_by" not in update_fields:
                kwargs["update_fields"] = set(update_fields) | {"modified_by"}

        return super().save(*args, **kwargs)



from django.contrib import admin

from django_rubble.models.number_models import NamedSerialNumber


@admin.register(NamedSerialNumber)
class NamedSerialNumberAdmin(admin.ModelAdmin):
    list_display = ["name", "content_type", "next_number"]
    fields = ["name", "content_type", "next_counter", "step", "prefix", "width"]



from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _
from loguru import logger
from pydantic import BaseModel


class SerialNumberConfig(BaseModel):
    """Configuration for NumberModel numbering."""

    name: str | None = None
    initial_value: int = 1
    step: int = 1
    prefix: str
    width: int


class NaturalKeyModelManager(models.Manager):
    def get_by_natural_key(self, *args, **kwargs):
        return self.get(*args, **kwargs)


class NaturalKeyModel(models.Model):
    natural_key_fields: list[str]

    objects = NaturalKeyModelManager()

    class Meta:
        abstract = True

    def natural_key(self):
        return (getattr(self, field_name) for field_name in self.natural_key_fields)


class NamedSerialNumberManager(models.Manager):
    def get_serial_number(
        self,
        parent_model,
        serial_number_config: SerialNumberConfig | None = None,
    ) -> tuple[str, bool]:
        """Generate next serial number, creating NamedSerialModel if one doesn't exist."""  # noqa: E501
        if serial_number_config is None:
            serial_number_config = parent_model.number_config

        model_type = ContentType.objects.get_for_model(parent_model)
        serial_number_config.name = (
            model_type.name
            if serial_number_config.name is None
            else serial_number_config.name
        )

        msg = f"SerialNumberConfig: {serial_number_config}"
        logger.debug(msg)

        named_serial, created = self.get_or_create(
            content_type=model_type,
            defaults={
                "name": serial_number_config.name,
                "next_counter": serial_number_config.initial_value,
                "prefix": serial_number_config.prefix,
                "width": serial_number_config.width,
                "step": serial_number_config.step,
            },
        )
        return named_serial.get_next_number(), created  # type: ignore[attr-defined]


class NamedSerialNumber(models.Model):
    name = models.CharField(_("name"), max_length=50, blank=True)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    next_counter = models.IntegerField(_("next counter"), default=1)
    step = models.IntegerField(_("step"), default=1)
    prefix = models.CharField(_("prefix"), max_length=10)
    width = models.IntegerField(_("number width"))

    objects = NamedSerialNumberManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["name", "content_type"], name="unique_name_for_content_type"
            )
        ]

    def __str__(self) -> str:
        return f"{self.name} -> next: {self.next_number}"

    @property
    @admin.display
    def next_number(self) -> str:
        return f"{self.prefix}{str(self.next_counter).zfill(self.width)}"

    def get_next_number(self) -> str:
        """Get current formatted number and increment counter."""
        next_number = self.next_number
        self.next_counter += self.step
        self.save()

        return next_number


class NumberedModelManager(models.Manager):
    def get_by_natural_key(self, number: str):
        return self.get(number=number)


class NumberedModel(models.Model):
    """Adds a `number` field that uses `DocumentNumber` to generate values.

    Args:
        number_config: A dictionary with the following key:value pairs:
            "prefix": str,
            "width": int,
            "start_value": 1
        number_prefix: A string prefix, case-sensitive
        number_width: An int representing how wide the numeric part of the string
        should be.
        number_start_value: An int used to set initial value. Only used for
        initial creation.


    Example:
        Assuming `number_prefix = "INV"` and `number_width = 4`, `INV0001`,
        `INV0002`, etc.
    """

    number = models.CharField(_("number"), unique=True, max_length=10, editable=False)
    number_config: SerialNumberConfig

    class Meta:
        abstract = True

    def save(self, *args, **kwargs) -> None:
        if self._state.adding:
            self.number, _ = NamedSerialNumber.objects.get_serial_number(self)
            msg = f"New number is: {self.number}"
            logger.debug(msg)
        return super().save(*args, **kwargs)

    def natural_key(self):
        return (self.number,)



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from django_rubble.utils.numbers import Percent, is_number


class SimplePercentageField(forms.DecimalField):
    def __init__(
        self,
        *,
        max_value=None,
        min_value=None,
        max_digits=None,
        decimal_places=None,
        **kwargs,
    ):
        if "widget" not in kwargs:
            step = 10 ** (-1 * decimal_places)
            kwargs["widget"] = forms.NumberInput(
                attrs={"class": "percent", "step": step}
            )
        self.max_digits, self.decimal_places = max_digits, decimal_places
        super().__init__(max_value=max_value, min_value=min_value, **kwargs)

    def to_python(self, value):
        val = super().to_python(value)

        if isinstance(val, Percent):
            return val.value

        if is_number(val):
            new_val = Percent.fromform(val)

            return new_val.value

        rtype = type(val)
        raise ValidationError(
            _("Invalid value type: %(rtype)s"),
            code="invalid",
            params={"rtype": rtype},
        )

    def prepare_value(self, value):
        val = super().prepare_value(value)

        if isinstance(val, Percent):
            return val.per_hundred
        if is_number(val):
            if isinstance(val, str):
                new_val = Percent.fromform(val)

                return new_val.per_hundred

            return Percent(value=val).per_hundred

        return val



from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic import CreateView, UpdateView


class CreateStampedView(LoginRequiredMixin, CreateView):
    """A view that creates a StampedModel instance."""

    def form_valid(self, form):
        form.instance.created_by = self.request.user
        form.instance.modified_by = self.request.user
        return super().form_valid(form)


class UpdateStampedView(LoginRequiredMixin, UpdateView):
    """A view that updates a StampedModel instance."""

    def form_valid(self, form):
        form.instance.modified_by = self.request.user
        return super().form_valid(form)



from django.contrib import admin  # type: ignore[import-untyped]


class StampedAdmin(admin.ModelAdmin):
    """A ModelAdmin that sets the created_by and modified_by fields of a StampedModel
    instance to the user that created or modified the instance.
    """

    def save_model(self, request, obj, form, change):
        if not change:
            obj.created_by = request.user
        obj.modified_by = request.user
        super().save_model(request, obj, form, change)


class StampedTabularInline(admin.TabularInline):
    """An InlineModelAdmin that sets the created_by and modified_by fields of a
    StampedModel instance to the user that created or modified the instance.
    """

    def save_model(self, request, obj, form, change):
        if not change:
            obj.created_by = request.user
        obj.modified_by = request.user
        super().save_model(request, obj, form, change)



from decimal import Decimal
from typing import Any

from pydantic import BaseModel


def is_number(s: Any):
    """Check if a value can be coerced into a number type."""
    if s is None:
        return False
    try:
        float(s)
    except ValueError:
        return False
    else:
        return True


def ratio_to_whole(ratio: Decimal | float | str) -> Decimal:
    """Convert a ratio to a whole number.

    This is useful for converting a ratio to a percentage.

    For example, 0.03 would be converted to 3 (3%).

    Args:
        ratio (Decimal, float, str): The ratio to be converted.

    Returns:
        The whole number. Decimal
    """
    return Decimal(str(ratio)) * Decimal("100")


def whole_to_ratio(whole: Decimal | float | str) -> Decimal:
    """Convert a whole number to a ratio.

    This is useful for converting a percentage to a ratio.

    For example, 3 would be converted to 0.03 (3%).

    Args:
        whole (Decimal, float, str): The whole number to be converted.

    Returns:
        The ratio. Decimal
    """
    return Decimal(str(whole)) * Decimal("0.01")


def trim_trailing_zeros(value: float | Decimal | str) -> Decimal:
    """Remove trailing zeros from a decimal value.

    This is useful for ensuring that a value can be safely compared with another value.

    For example, 3.1400 would be displayed as 3.14.

    Args:
        value (float, Decimal, str): The value to be trimmed.

    Returns:
        The trimmed value. Decimal
    """
    return Decimal(str(value)).normalize()


def set_zero(value: float | Decimal | str) -> Decimal:
    """Set a value to a true Decimal zero if it is zero."""
    decimal_from_string = Decimal(str(value))

    if decimal_from_string == Decimal(0):
        return Decimal()

    return decimal_from_string


class Percent(BaseModel):
    value: Decimal | float | str
    per_hundred: Decimal | float | str | None = None
    decimal_places: int | None = None
    has_decimal_places: bool | None = None

    def model_post_init(self, __context: Any) -> None:
        new_value = trim_trailing_zeros(self.value)
        per_hundred_dec = trim_trailing_zeros(ratio_to_whole(self.value))

        if self.decimal_places is not None:
            new_value = round(new_value, self.decimal_places + 2)
            per_hundred_dec = round(per_hundred_dec, self.decimal_places)
            self.has_decimal_places = True
        else:
            self.has_decimal_places = False

        self.value = set_zero(new_value)
        self.per_hundred = set_zero(per_hundred_dec)

        super().model_post_init(__context)

    @classmethod
    def fromform(cls, val: Decimal, field_decimal_places: int | None = None):
        """Create Percent from human-entry (out of 100)"""
        ratio_decimal = whole_to_ratio(val)
        return cls(value=ratio_decimal, decimal_places=field_decimal_places)

    def __mul__(self, other):
        """Multiply using the ratio (out of 1) instead of human-readable out of 100"""
        return self.value.__mul__(other)

    def __float__(self):
        return float(self.value)

    def as_tuple(self):
        return self.value.as_tuple()

    def is_finite(self):
        return self.value.is_finite()

    def __repr__(self) -> str:
        return f"Percentage('{self.value}', '{self.per_hundred}%')"

    def __str__(self):
        return f"{self.per_hundred}%"
